/* Copyright 2000 Servaas Vandenberghe, Leuven, Belgium, All rights reserved.
 This software is licensed under the GNU GENERAL PUBLIC LICENSE, version 2, or
 a later version at your option.

 Redistribution and use of this source code, with or without modification, is
 permitted provided that the following condition is met:

 1. Redistributions of this source code must retain the above copyright
    notice, this condition, and the following disclaimer.

  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO
  EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <errno.h> 
#include <limits.h>  /* INT_MAX */
#include <stdlib.h>  /* strtol */
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>   /* getopt */
#include <stdio.h>
#include <string.h>

#ifndef VERBOSE
#define VERBOSE 0
#endif

#ifndef VERSION
#define VERSION "0.0"
#endif
static char version[]= VERSION ", " __DATE__ ", " __TIME__ ;

static char progname[1025]="getdsr";
static const char *cname="getdsr";

static void usage(void ) {
  printf("usage: %s [-DdhvV] [-s <int>] tty prog args \n", progname);
  printf("Try `%s -h' for more information \n", progname);
}

static void help(void) {
  printf("usage: %s [-DdhVv] [-s <int>] tty prog args \n", progname);
  printf("%s %s \n", cname, version);
  printf("%s waits for a high DSR and executes prog \n",cname);
  printf("  -D        disable the DTR line \n"
	 "  -d        enable the DTR line \n"
	 "  -s <int>  sleep time before rechecking the DSR line \n"
	 "  -V        version   \n"
	 "  -v        verbose   \n"
	 );
}

/* This routine opens the TTY and reads TIOCMGET. */
#define WHERE "main"
int main(int argc, char *argv[]) {
  const char *where=cname; /* []=WHERE; */
  int err=0, verbose=0;
  char *ttyfile=NULL, buf[258];
  int ttylen;
  int ttyfd=0, ttyclose=0;
  int errcm, ttycm;
  int ttysleep=30;
  int odtrs=0, odtrc=0;
  int opt;
  extern char *optarg;         /* pointer to the argument of the option */
  extern int optind, optopt;   /* opterr; */

  if(argc) {
    strncpy(progname,argv[0], sizeof(progname)-1);
    progname[sizeof(progname)-1] = '\0';
  }
  if(argc <= 1) {
    usage();
    err=1; goto end_error;
  }

  optarg = NULL;
  optind = 0;
  while( (opt = getopt(argc, argv, ":dDhl:s:vV")) != EOF ) {
    int this_optind = optind-1;
    switch(opt) {
    case 'D':
      odtrc ^= 1;
      break;
    case 'd':
      odtrs ^= 1;
      break;
    case 'h':
      help(); 
      goto end_error;
    case 'l':
      ttyfile = optarg;
      break;
    case 's': 
      { long val;
        char *endptr;

	val = strtol(optarg, &endptr, 0);
	if (*optarg == '\0' || *endptr != '\0') {
	  fprintf(stderr, "%s: failed to extract an integer from \"%.99s\"\n", 
		  where, optarg);
	  err = 5; goto end_error;
	}
	if (val < 0 || val > INT_MAX) {
	  fprintf(stderr, "%s: bad range integer from \"%.99s\"\n", 
		  where, optarg);
	  err = 6; goto end_error;
	}
	ttysleep = val;
      }
      break;
    case 'v':
      verbose++; 
      break;
    case 'V':
      printf("%s %s\n", cname, version); 
      goto end_error;
    case ':':
      fprintf(stderr, "%s: missing option argument at %i: -%c\n", where, 
	      this_optind, optopt);
      err = 7; goto end_error;	
    case '?':
      fprintf(stderr, "%s: error getopt unknown option %i: -%c\n", where, 
	      this_optind, optopt);
      err = 8; goto end_error;
    default:
      fprintf(stderr, "%s: ?? getopt returned character code 0x%x\n", where, 
	      opt);
      err = 9; goto end_error;
    }
  }

  /* option "-l ttyS0" not present => check for tty on command line */
  if(!ttyfile && optind < argc)
    ttyfile=argv[optind++];

  if (!ttyfile) {
    help(); 
    goto end_error;
  }
  /* progname must follow */
  if (optind >= argc) {
    fprintf(stderr, "%s: program to execute not specified\n", where);
    err=10; goto end_error;
  }

  /* prepend /dev/ if not an absolute pathname */
  ttylen = strlen(ttyfile);
  if (ttylen && (ttyfile[0]!='-' || ttylen>1)) {
    if (ttyfile[0] != '/') {
      snprintf(buf, sizeof(buf)-1, "/dev/%s", ttyfile);
      buf[sizeof(buf)-1] = '\0';
      ttyfile = buf;
      ttylen = strlen(ttyfile);
    }
  }

  /**** normal operation below ****/
  for (;;) {
    ttycm = 0xFFFF;  /* test all flags 1 */
    errcm = 0;       /* last ioctl() return */

    /** open the tty if not stdin **/
    if (ttylen && (ttyfile[0]!='-' || ttylen>1)) {
      int fatal;

      for (fatal = 0; !fatal; ) {            /* try forever open() */
	ttyfd = open(ttyfile, O_NONBLOCK);   /* proceed even if DSR = 0 */
	if (ttyfd < 0 && verbose)
	  fprintf(stderr, "%s: bad fd=%i open(%s) %i-%s\n", where, ttyfd,
		  ttyfile, errno, strerror(errno));

	if (!ttysleep || ttyfd >= 0)
	  break;

	sleep(ttysleep);
#ifdef ENOMEM
	if (errno == ENOMEM)
	  fatal |= 1;
#endif
#ifdef EMFILE
	if (errno == EMFILE)
	  fatal |= 2;
#endif
#ifdef ENFILE
	if (errno == ENFILE)
	  fatal |= 4;
#endif
      } /* for !fatal */
      /* !ttysleep || fatal || ttyfd>=0 */
      if (ttyfd < 0) {
	err = 21;
	if (verbose)
	  fprintf(stderr, "%s: error open() return code %i\n", where, err);
	goto end_error;
      }
      ttyclose = 1;

      if (verbose > 1)
	printf("%s: open(%s) fd=%i\n", where, ttyfile, ttyfd);
    } /* if ttylen */

    /** clear and/or set DTR line.  Most OSes set DTR on open() **/
    if (odtrc || odtrs) {
      int cm, ret;

      errcm = ioctl(ttyfd, TIOCMGET, &cm);
      if (errcm < 0) {
	if (verbose)
	  fprintf(stderr, "%s: DTR bad errcm=%i ioctl(%i,TIOCMGET) %i-%s\n",
		  where, errcm, ttyfd, errno, strerror(errno));
      } else {
	int i;
	for (i = 0; i < 2; i++) {
	  if (!i) {
	    if(odtrc)
	      cm &= ~TIOCM_DTR; 
	    else
	      continue;
	  } else {
	    if(odtrs) 
	      cm |=  TIOCM_DTR; 
	    else
	      continue;
	  }
	  ret = ioctl(ttyfd, TIOCMSET, &cm);
	  if (ret && verbose)
	    fprintf(stderr, "%s: bad ret=%i ioctl(%i,TIOCMSET) %i-%s\n", where,
		    ret, ttyfd, errno, strerror(errno));
	  sleep(1);
	} /* for i */
      }
    } else
      if (ttyclose)  /* just opened the line: wait for the modem */
	sleep(1);

    /* get the modem status lines
     * EINVAL for tty/pty pairs
     * EIO    device node open but hardware not responsive
     */
    if (errcm >= 0) {
      for (;;) {                             /* try forever ioctl() */
	errcm = ioctl(ttyfd, TIOCMGET, &ttycm);

	if(verbose) {
#define MSGS 9
	  int listcm[MSGS] = {
	    TIOCM_LE, TIOCM_DTR, 
	    TIOCM_RTS, TIOCM_ST, 
	    TIOCM_SR, TIOCM_CTS, 
	    TIOCM_CAR, TIOCM_RNG, 
	    TIOCM_DSR 
	  };
	  char *listmsg[MSGS+1] = {
	    "line enable", "data terminal ready",
	    "request to send", "secondary transmit",
	    "secondary receive", "clear to send",
	    "carrier detect", "ring",
	    "data set ready", NULL
	  };
#undef MSGS
	  int i;

	  if (errcm < 0)
	    fprintf(stderr, "%s: bad errcm=%i ioctl(%i,TIOCMGET) %i-%s\n",
		    where, errcm, ttyfd, errno, strerror(errno));

	  printf("errcm=%i ttycm=%#04x\n", errcm, ttycm);
	  for(i = 0; listmsg[i]; i++)
	    printf("%c %s\n", ttycm & listcm[i] ? '1' : '0', listmsg[i]);
	} /* if verbose */

	if (!ttysleep || errcm < 0 || (ttycm & TIOCM_DSR))
	  break;
	if (verbose)
	  printf("%s: ioctl() sleeping %i s\n", where, ttysleep);
	sleep(ttysleep);
      } /* for ever */
    } /* if modem control */

    /* close, all done or will reopen on error */
    if (ttyclose) {
      int ret = close(ttyfd);
      if (ret < 0 && verbose)
	fprintf(stderr, "%s: bad ret=%i close(%i) %i-%s\n", where, ret, 
		ttyfd, errno, strerror(errno));
      ttyclose = 0;
    }

    if (!ttysleep || errcm >= 0)
      break;
    if (verbose)
      printf("%s: retry in %i s\n", where, ttysleep);
    sleep(ttysleep);
  } /* for ever main loop */

  /* out of loop: !ttysleep || errcm >= 0 */
  if (errcm < 0) {
    err = 25;
    if (verbose)
      fprintf(stderr, "%s: error aborting with return code %i\n", where, err);
    goto end_error;
  }

  /* execute the program */
  if (ttycm & TIOCM_DSR) {
    int filelen;
    filelen = strlen(argv[optind]);
    if(filelen && (argv[optind][0] != '-' || filelen > 1)) {
      int ret;
      ret = execvp(argv[optind], &argv[optind]);
      if (ret) {
	fprintf(stderr, "%s: error %i execvp %i-%s", where, ret,
		errno, strerror(errno));
	err = 27; goto end_error;
      }
    } /* if filelen */
  } else {
    err = 29;
    if(verbose)
      fprintf(stderr, "%s: error no DSR, return code %i\n", where, err);    
  }

end_error:
  if (ttyclose) {
    int ret;
    fprintf(stderr, "%s: error program ttyclose=%i ttyfd=%i\n", where, 
	    ttyclose, ttyfd); 
    ret = close(ttyfd);
    if(ret) {
      fprintf(stderr, "%s: error %i close(%i) %i-%s\n", where, ret, ttyfd, 
	      errno, strerror(errno));
      if(!err) err=98;
    }
  }
  return err;
}
#undef WHERE

#undef VERBOSE
#undef VERSION
